# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Validate by confirming no duplicate items."""

import argparse
import pathlib
from abc import ABC, abstractmethod

from trestle.core.base_model import OscalBaseModel
from trestle.core.validator_helper import find_values_by_name
from trestle.utils import fs


class Validator(ABC):
    """Abstract Validator interface."""

    @abstractmethod
    def validate(self, args: argparse.Namespace) -> int:
        """Validate the model."""


class DuplicatesValidator(Validator):
    """Find duplicate items in oscal object."""

    def validate(self, args: argparse.Namespace) -> int:
        """Perform the validation."""
        file_path = pathlib.Path(args.file).absolute()
        model_type, _ = fs.get_contextual_model_type(file_path)
        model: OscalBaseModel = model_type.oscal_read(file_path)

        loe = find_values_by_name(model, args.item)
        if loe:
            nitems = len(loe)
            return 0 if nitems == len(set(loe)) else 1
        return 0
